from django.conf import settings
from django.db import connections, connection
from django.db.backends.utils import logger as db_logger
from django.utils.encoding import force_str, force_unicode

import flatdict
import types
from time import time

from .exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from .utils import prettify


class DictCursorWrapper(object):
    def __init__(self, cursor):
        self._cursor = cursor

    def __getattr__(self, attr):
        return getattr(self._cursor, attr)

    def dictfetchall(self):
        columns = [col[0] for col in self._cursor.description]
        return [
            dict(zip(columns, row))
            for row in self._cursor.fetchall()
            ]


class SQLResult(object):
    def __init__(self, cursor):
        self._cursor = cursor

    @property
    def cursor(self):
        return self._cursor


class SQLValuesResult(SQLResult):
    def __init__(self, *args, **kw):
        super(SQLValuesResult, self).__init__(*args, **kw)
        self._result = None
        self._columns = [col[0] for col in self._cursor.description]

    @property
    def result(self):
        if self._result is None:
            self._result = self._fetch(self._cursor)
        return self._result

    def __iter__(self):
        return iter(self.result)

    def _fetch(self, cursor):
        raise NotImplementedError

    def __str__(self):
        return str(self.result)

    def __getitem__(self, idx):
        return self.result[idx]


class ValuesListResult(SQLValuesResult):
    def _fetch(self, cursor):
        return cursor.fetchall()

    def __len__(self):
        return len(self.result)


class ValuesResult(SQLValuesResult):
    def _fetch(self, cursor):
        return [
            dict(zip(self._columns, row))
            for row in cursor.fetchall()
            ]

    def __len__(self):
        return len(self.result)


class ValuesListIterator(SQLValuesResult):
    def _fetch(self, cursor):
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            yield row


class ValuesIterator(SQLValuesResult):
    def _fetch(self, cursor):
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            yield dict(zip(self._columns, row))


class TemplateQuery(object):
    def __init__(self, template, context=None, using=None):
        self._template = template
        self._context = context or {}
        self._using = using
        self._sql = None

    @property
    def sql(self, using=None):
        if self._sql is None:
            self._sql = self.render()
        return self._sql

    def render(self):
        return self._template.render(self._context)

    @property
    def context(self):
        return self._context

    def bind(self, **context):
        ctx = dict(self._context)
        ctx.update(context)
        return self._clone(context=ctx)

    def using(self, using):
        return self._clone(using=using)

    def _cursor(self, using=None):
        using = using or self._using
        conn = connections[using] if using else connection
        return DictCursorWrapper(conn.cursor())

    def execute(self, using=None, result_class=SQLResult):
        def extract_subcontexts(values):
            subcontexts = []
            for value in values:
                try:
                    subctx = value.context
                except AttributeError:
                    pass
                else:
                    subcontexts.append(subctx)
                    subcontexts+= extract_subcontexts(subctx.values())
            return subcontexts

        ctx = {}

        for subcontext in extract_subcontexts(self._context.values()):
            ctx.update(subcontext)
        ctx.update(self._context)

        query_params = flatdict.FlatDict(ctx, delimiter='.').as_dict()

        cur = self._cursor(using=using or self._using)

        # Get native cursor to avoid problems with Django` DebugCursor
        # logging. See https://code.djangoproject.com/ticket/22377

        if settings.DEBUG:
            start = time()

        native_cursor = cur.cursor
        native_cursor.execute(force_str(self.sql), query_params)

        # Until Django bug #22377 is not resolved, we must manually
        # log query and append SQL to queries_log

        if settings.DEBUG:
            stop = time()
            duration = stop-start

            sql = cur.db.ops.last_executed_query(
                    native_cursor, self.sql, query_params)

            cur.db.queries_log.append({
                'sql': sql,
                'time': "%.3f" % duration,
            })

            db_logger.debug(
                '(%.3f) %s; args=%s', duration, self.sql, query_params,
                extra={'duration': duration, 'sql': self.sql, 'params': query_params}
            )

        return result_class(cur)

    def values(self, **extra):
        obj = self.bind(**extra) if extra else self
        return obj.execute(result_class=ValuesResult)

    def values_list(self, **extra):
        obj = self.bind(**extra) if extra else self
        return obj.execute(result_class=ValuesListResult)

    def iterator(self, **extra):
        obj = self.bind(**extra) if extra else self
        return obj.execute(result_class=ValuesListIterator)

    def dictiterator(self, **extra):
        obj = self.bind(**extra) if extra else self
        return obj.execute(result_class=ValuesIterator)

    def scalar(self, **extra):
        obj = self.bind(**extra) if extra else self
        result = obj.values_list(**extra)

        try:
            row = result[0]
        except IndexError:
            raise ObjectDoesNotExist('Query returned no rows')

        if len(result)>1:
            raise MultipleObjectsReturned('Query returned more than one row')

        return row[0]

    def rawtuple(self):
        return self.sql, self.context

    def pretty(self):
        return prettify(self.sql)

    def _clone(self, **kwargs):
        template = kwargs.pop('template', self._template)
        context = kwargs.pop('context', self._context)
        using = kwargs.pop('using', self._using)

        return type(self)(
            template=template, context=context, using=using, **kwargs)

    def __unicode__(self):
        return force_unicode(self.sql)

    def __str__(self):
        return force_str(unicode(self))



from django.template.utils import get_app_template_dirs
from django.template.loaders.filesystem import Loader as FilesystemLoader
from django.core.exceptions import SuspiciousFileOperation
from django.utils._os import safe_join

try:
    from django.template import Origin # Django 1.9
except ImportError:
    Origin = None


class Loader(FilesystemLoader):
    is_usable = True

    def get_template_sources(self, template_name, template_dirs=None):
        """
        Returns the absolute paths to "template_name", when appended to each
        directory in "template_dirs". Any paths that don't lie inside one of the
        template dirs are excluded from the result set, for security reasons.
        """
        if not template_dirs:
            template_dirs = self.get_dirs()
        for template_dir in template_dirs:
            try:
                name = safe_join(template_dir, template_name)
            except SuspiciousFileOperation:
                # The joined path was located outside of this template_dir
                # (it might be inside another one, so this isn't fatal).
                pass
            else:
                if Origin:
                    yield Origin(
                        name=name,
                        template_name=template_name,
                        loader=self,
                    )
                else:
                    yield name

    def get_dirs(self):
        return get_app_template_dirs('sqltemplates')




from django.template.utils import EngineHandler
from .settings import TEMPLATES
import contextlib


engines = EngineHandler(templates=TEMPLATES)


class SQLTemplate(object):
    def __init__(self, using=None, context=None):
        self._using = using
        self._context = context or {}

    def get(self, template_name, context=None, using=None):
        from .loader import get_template
        from .query import TemplateQuery

        ctx = dict(self._context)
        ctx.update(context or {})

        return TemplateQuery(
                template=get_template(template_name),
                context=ctx, using=using or self._using
                )

    def from_string(self, string, context=None, using=None):
        from .query import TemplateQuery

        default_engine = engines.all()[0]

        ctx = dict(self._context)
        ctx.update(context or {})

        return TemplateQuery(
                template=default_engine.from_string(string),
                context=ctx, using=using or self._using
                )


service = SQLTemplate()


def get(*args, **kw):
    return service.get(*args, **kw)


def from_string(*args, **kw):
    return service.from_string(*args, **kw)


@contextlib.contextmanager
def context(**context):
    yield SQLTemplate(context=context)


@contextlib.contextmanager
def using(name):
    yield SQLTemplate(using=name)


@contextlib.contextmanager
def scope(context=None, using=None):
    yield SQLTemplate(using=using, context=context)




from django.conf import settings


TEMPLATES = getattr(settings, 'SQL_TEMPLATES', [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': [],
    'APP_DIRS': False,
    'OPTIONS': {
        'loaders': [
            'django.template.loaders.filesystem.Loader',
            'sqltemplate.loaders.app_directories.Loader',
            ]
        },
    }])

